// Message.java
//	Class for distributed filesystem (DFS) messages.
//
// Copyright (c) 2005 State University of New York at Stony Brook.
// All rights reserved.  See the COPYRIGHT file for copyright notice and 
// limitation of liability and disclaimer of warranty provisions.

package nachos.kernel.filesys;

import java.io.*;

import nachos.Debug;

/**
 * Class for messages representing a request to or a reply from
 * a distributed filesystem (DFS) server.  */
public class Message {

  /**
   *  A type of DFS message, used as a value of field msgType.
   */
  public static final int OpenReq = 0, CloseReq = 1, ReadReq = 2, 
    WriteReq = 3, CreateReq = 4, RemoveReq = 5, OpenReply = 6, 
    ReadReply = 7, WriteReply = 8;

  /** Names of the types of DFS messages (useful for debugging). */
  public static final String msgTypeNames[] = { "OpenReq", "CloseReq", "ReadReq", "WriteReq", "CreateReq", "RemoveReq", "OpenReply", "ReadReply", "WriteReply" };

  /** type of message. */
  public int msgType;

  /** The file descriptor.  This field is meaningful iff msgType is
   * CloseReq, ReadReq, WriteReq, or OpenReply. */
  public int fileDesc;

  /** The starting position within the file.  This field is meaningful
   * iff msgType is ReadReq, WriteReq.  */
  public int position;

  /** The number of bytes to be read, or the number of bytes
   * successfully written.  This field is meaningful iff msgType is
   * ReadReq or WriteReply.  */
  public int numBytes;

  /** The file name.  This field is meaningful iff msgType is OpenReq,
   * CreateReq, or RemoveReq.  Otherwise, filename should be null. */
  public String filename;

  /** The file data to write or read.  This field is meaningful iff
   * msgType is WriteReq or ReadReply.  If msgType is not WriteReq or
   * ReadReply, data should be null. */
  public byte[] data;

  /** Length of the message header, in bytes.  The header consists of
   * all fields except filename and data.  Each field occupies 1 byte,
   * except the position field, which occupies 2 bytes. */
  public static final int headerLength = 5;

  /** The size of data blocks used in the distributed filesystem.  To ensure
   * that a Message containing a block of data will fit in packet, we
   * require BlockSize <= Packet.maxContentsLength - Message.headerLength. */
  public static final int BlockSize = 32;

  static {
    Debug.ASSERT(BlockSize <= (nachos.machine.Packet.maxContentsLength - headerLength));
  }

  /**
   * Allocate a new message, using the specified parameters, which
   * have the same meaning as the fields with the same names.
   */
  public Message(int msgType, int fileDesc, int position, int numBytes, String filename, byte[] data) {
	    
	this.msgType = msgType;
	this.fileDesc = fileDesc;
	this.position = position;
	this.numBytes = numBytes;
	this.filename = filename;
	this.data = data;
	Debug.ASSERT(0 <= msgType && msgType <= 8);
	Debug.ASSERT(0 <= fileDesc && fileDesc <= Byte.MAX_VALUE);
	Debug.ASSERT(0 <= position && position <= Byte.MAX_VALUE * Byte.MAX_VALUE);
	// check that the requested data, filename, or supplied data,
	// respectively, have size at most BlockSize, to ensure it
	// (together with the message header) will fit in a packet.
	Debug.ASSERT(numBytes <= BlockSize);
	Debug.ASSERT(filename==null || filename.getBytes().length <= BlockSize);
	Debug.ASSERT(data == null || data.length <= BlockSize);
  }

  /** Returns this message represented as an array of bytes.
   * @return this message, represented as an array of bytes. */
  public byte[] toByteArray() {
    if (msgType == OpenReq || msgType == CreateReq || msgType == RemoveReq) {
      // temporarily store the filename, encoded as a byte array, in this.data.
      data = filename.getBytes();
    }
    int length = headerLength + data.length;
    byte[] result = new byte[length];
    result[0] = (byte)msgType;
    result[1] = (byte)fileDesc;
    result[2] = (byte)(position / Byte.MAX_VALUE);
    result[3] = (byte)(position % Byte.MAX_VALUE);
    result[4] = (byte)numBytes;
    System.arraycopy(data, 0, result, headerLength, data.length);
    if (msgType == OpenReq || msgType == CreateReq || msgType == RemoveReq) {
      data = null;
    }
    return result;

    // just for fun, here is an easier, but less efficient, approach.
    //ByteArrayOutputStream os = new ByteArrayOutputStream();
    //ObjectOutputStream oos = new ObjectOutputStream(os);
    //oos.writeObject(this);
    //result = os.toByteArray();
  }

  /**
   * Create a new message, and initialize it using the specified array
   * of bytes, which should have been produced by Message.toByteArray.
   *
   * @param msgAsBytes  the message, represented as an array of bytes.  */
  public Message(byte[] msgAsBytes) {
    msgType  = msgAsBytes[0];
    fileDesc = msgAsBytes[1];
    position = ((int)msgAsBytes[2])*Byte.MAX_VALUE + (int)msgAsBytes[3];
    numBytes = msgAsBytes[4];
    int dataLen = msgAsBytes.length - headerLength;
    if (msgType == OpenReq || msgType == CreateReq || msgType == RemoveReq) {
      filename = new String(msgAsBytes, headerLength, dataLen);
    }
    if (msgType == WriteReq || msgType == ReadReply) {
      data = new byte[dataLen];
      System.arraycopy(msgAsBytes, headerLength, data, 0, dataLen);
    }

    // here is the other half of the easier but less efficient approach.
    //ByteArrayInputStream is = new ByteArrayInputStream(msgAsBytes);
    //ObjectInputStream ois = new ObjectInputStream(is);
    //return (Message)ois.readObject();
  }

  /** Returns this message represented as a string.  Useful for debugging.
   * @return this message, represented as a string. */
  public String toString() {
    StringBuffer b = new StringBuffer("msgType: "+msgType+", "+
				      "filedesc: "+fileDesc+", "+
				      "position: "+position+", "+
				      "numBytes: "+numBytes+", "+
				      "filename: "+filename+", "+
				      "data: ");
    if (data != null) for (int i=0; i<data.length; i++) b.append(data[i]+" ");
    return b.toString();
  }

}
